from incywincy.visitors import http_errors, absolute_links


# user, password = 'user', 'password'


def nature(page):
    print('Nature! [{0}]'.format(page.url))


visitors = (
    ('.*', http_errors),
    ('.*', absolute_links),
    (r'^/nature', nature),
)



import re
from urlparse import urljoin
from collections import deque

from bs4 import BeautifulSoup
import requests
from requests.auth import HTTPBasicAuth

from incywincy.config import settings
from incywincy.report import debug, error


UA = 'Mozilla/5.0 (X11; U; Linux i686) Gecko/20071127 Firefox/2.0.0.11'
exceptions = []


def find_links(html):
    """
    Find all ``<a>`` links in the HTML page.
    """
    soup = BeautifulSoup(html)
    links = filter(lambda a: 'href' in a.attrs, soup.find_all('a'))
    return [a['href'] for a in links]


def normalised(links, start, base):
    """
    Return a set of normalised links.

    These will all be absolute URLs, and no repeats.
    """
    links = [urljoin(base, link) for link in links]
    return set(filter(lambda x: x.startswith(start), links))


class Page(object):
    """
    A web page, that can be processed by a visitor.

    Contains useful information about the page, such as the status code,
    content, list of links, and referrer (the page where we found a link to
    this one).
    """
    def __init__(self, url, referrer):
        self.referrer = referrer
        self.url = url
        headers = {'User-Agent': UA}
        self.response = requests.get(url, auth=settings.auth, headers=headers)
        self._links = []

    def __getattr__(self, attr):
        return getattr(self.response, attr)

    def __str__(self):
        if isinstance(self.referrer, Page):
            referrer_url = self.referrer.url
        else:
            referrer_url = self.referrer
        return "{url} [from: {ref}]".format(url=self.url, ref=referrer_url)

    @property
    def links(self):
        if not self._links:
            if self.response.status_code == 200:
                content_type = self.headers['Content-Type']
                if content_type.startswith('text/html'):
                    self._links = find_links(self.text)
                    # debug('Links: ' + ', '.join(self._links))
        return self._links

    def normalised_links(self, start):
        return normalised(self.links, start, self.url)


def visit(url, parent=None):
    """
    Recursively visit all pages from this URL.

    Will keep track of seen pages, so as to avoid processing links twice.
    """
    try:
        settings.auth = HTTPBasicAuth(settings.user, settings.password)
    except:
        settings.auth = None

    # Keep a queue of pending links, and a set of seen links
    pending = deque()
    seen = set()

    pending.append((url, parent))
    seen.add(url)
    total = 1

    # Main crawler loop
    while len(pending) > 0:
        url, parent = pending.popleft()
        debug('[{0},+{1}] Opening: {2}'.format(total, len(pending), url))
        try:
            page = Page(url, parent)
        except Exception as e:
            error('Exception in {0}: {1}'.format(url, e))
            exceptions.append({'url': url, 'error': e})
        else:
            for pattern, visitor in settings.visitors:
                url_path = url[len(settings.root):]
                if re.match(pattern, url_path):
                    visitor(page)
            # add new links to queue if successful
            if page.status_code == 200:
                links = page.normalised_links(settings.start)
                links.difference_update(seen)
                # debug("New links:\n" + "\n".join(links))
                for link in links:
                    pending.append((link, page))
                    seen.add(link)
                total += len(links)



"""Incy Wincy Spider.

Usage:
    spider.py [--level=<level>] <settings> <url> [<start>]
    spider.py -h | --help

Options:
    -h --help        Show this screen.
    --version        Show version.
    --level=<level>  Set logging level (0: Debug to 3: Error) - [default: 2].
"""
if __name__ == '__main__':
    from docopt import docopt
    from incywincy import __version__, report
    from incywincy.config import settings
    from incywincy.http import visit

    arguments = docopt(__doc__, version='Incy Wincy ' + __version__)
    # print(arguments)
    settings.load(arguments['<settings>'])
    report.THRESHOLD = int(arguments['--level'])

    settings.root = arguments['<url>']
    if settings.root.endswith('/'):
        settings.root = settings.root[:-1]

    settings.start = arguments['<start>']
    if settings.start is None:
        settings.start = settings.root

    visit(settings.start, settings.root)



from incywincy.config import settings
from incywincy.report import log, ERROR, WARNING


def http_errors(page):
    status = page.status_code
    if status >= 400:
        log(page, 'Status [{0}]'.format(status), ERROR)


def absolute_links(page):
    for link in page.links:
        if link.startswith(settings.root):
            log(page, 'Absolute internal URL [{0}]'.format(link), WARNING)


def redirects(page):
    if page.history:
        location = page.history[0].headers['location']
        log(page, 'Redirected to: {0}'.format(location), WARNING)



#!/usr/bin/env python
# # coding: utf-8

from setuptools import setup
from incywincy import __version__


setup(
    name='incywincy',
    description='Web Spider Testing Framework',
    long_description='A framework to easily run tests by spidering through a site',
    version=__version__,
    author='Carles Barrobés',
    author_email='carles@barrobes.com',
    url='https://github.com/txels/incywincy',
    packages=['incywincy'],
    install_requires=['beautifulsoup4', 'requests'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Testing',
    ],
)



class Settings(dict):
    """
    Application settings, that will typically be loaded from a module.
    """
    def __init__(self):
        self._mod = None

    def load(self, module):
        self.module = module
        if self._mod is None:
            self._mod = __import__(self.module, globals(), locals(), [], -1)
            for attr in dir(self._mod):
                if not attr.startswith('__'):
                    self[attr] = getattr(self._mod, attr)

    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value


settings = Settings()



DEBUG = 0
INFO = 1
WARNING = 2
ERROR = 3

CRITICALITY = {
    DEBUG: 'DEBUG',
    INFO: 'INFO',
    WARNING: 'WARNING',
    ERROR: 'ERROR',
}

THRESHOLD = 1


def log(page, message, level=WARNING):
    if level >= THRESHOLD:
        print(">> {0}: {1} | {2}".format(CRITICALITY[level], message, page.url))


def debug(message):
    if THRESHOLD == DEBUG:
        print(message)


def error(message):
    print(message)
